"Query service for natbib metadata"


import requests


class QueryBuilder:
    """Query builder for the metadata_query2 service.

    This service allows for querying the Norwegian National Bibliography using Marc 21 fields and values.

    Example of usage:
    builder = QueryBuilder()
    result = (builder.add_field("field1", "subfield1", "value1")
          .add_field("field2", "subfield2", "value2")
          .set_return_fields("returnField1", "returnField2", "returnField3")
          .set_limit(1000)
          .post())

    """

    endpoint = (
        "https://api.nb.no/dhlab/metadata_query2" 
    )

    def __init__(self):
        self.query = {"fields": [], "returnFields": [], "limit": None}

    def add_field(self, field, subfield=None, value=None):
        """Add a field constraint to the query.

        Should be field + value for marc21 without subfields (<11) and field + subfield + value for marc21 with subfields (>=11
        """
        field_entry = {"field": field}
        if subfield:
            field_entry["subfield"] = subfield
        if value:
            field_entry["value"] = value
        self.query["fields"].append(field_entry)
        return self

    def set_return_fields(self, *fields):
        "Specify which fields to return in the response."
        self.query["returnFields"] = list(fields)
        return self

    def set_limit(self, limit):
        "Specify the maximum number of records to return."
        self.query["limit"] = limit
        return self

    def build(self):
        "Build the query."
        # Remove optional keys if they are not set
        if not self.query["returnFields"]:
            del self.query["returnFields"]
        if self.query["limit"] is None:
            del self.query["limit"]
        return self.query

    def post(self):
        "Post the query to the service and return the response."
        constructed_query = self.build()
        response = requests.post(self.endpoint, json=constructed_query)
        return response.json()



from typing import Dict, Optional

from dhlab.text.corpus import Corpus


class CorpusCollection:
    """A class for handling a collection of corpora."""

    def __init__(self, corpora: Optional[Dict[str, Corpus]] = None):
        """Initialize the class with a dictionary of corpora."""
        self.corpora = corpora if corpora is not None else {}

    def __getitem__(self, key: str) -> Corpus:
        """Get a corpus by name."""
        return self.corpora[key]

    def __setitem__(self, key: str, value: Corpus):
        """Set a corpus by name."""
        self.corpora[key] = value

    def __repr__(self) -> str:
        """Print the names of the corpora."""
        return "\n".join(self.corpora.keys())

    def __iter__(self):
        """Iterate over the names of the corpora."""
        for c in self.corpora:
            yield c

    def __len__(self) -> int:
        """Return the number of corpora."""
        return len(self.corpora)

    def __contains__(self, key: str) -> bool:
        """Check if a corpus is in the collection."""
        return key in self.corpora

    def add(self, name: str, corpus: Corpus):
        """Add a corpus to the collection."""
        self.corpora[name] = corpus

    def remove(self, name: str):
        """Remove a corpus from the collection."""
        del self.corpora[name]

    def get(self, name: str) -> Corpus:
        """Get a corpus by name."""
        return self.corpora[name]

    def show_corpora(self) -> Dict[str, Corpus]:
        """Show the corpora in the collection."""
        return self.corpora

    def concat_corpora(self) -> Corpus:
        """Concatenate all corpora in the collection into a single corpus."""
        new_corpus = Corpus()
        for c in self.corpora.values():
            new_corpus += c
        return new_corpus



from typing import Union
from IPython.display import HTML
from PIL import Image

from dhlab.api.nb_search_api import iiif_manifest, super_search, total_search, load_picture


def pages(urn, scale=800):
    a = iiif_manifest(urn)
    return [page['images'][0]['resource']['@id'].replace(f"full/full/", "full/0,{s}/") for page in
            a['sequences'][0]['canvases']]


def get_url(urn, page=1, part=200):
    # urn as digit
    urn = "URN:NBN:no-nb_digibok_" + urn + '_{0:04d}'.format(page)
    print(urn)
    url = f"https://www.nb.no/services/image/resolver/{urn}/full/0,{part}/native.jpg"
    return url


def page_urn(urn, page=1):
    # urn as digit
    return "URN:NBN:no-nb_digibok_" + urn + '_{0:04d}'.format(page)


def find_urls(term, number=50, page=0, mediatype='bilder'):
    """generates urls from super_search for pictures"""
    x = super_search(term, number, page, mediatype=mediatype)
    try:
        urls = [
            f['_links']['thumbnail_custom']['href']
            for f in x['_embedded']['items']
            if f['accessInfo']['accessAllowedFrom'] == 'EVERYWHERE'
               and 'thumbnail_custom' in f['_links']
        ]
    except:
        urls = []
    return urls


def find_urls2(term, number=50, page=0):
    """generates urls from super_search for pictures"""
    x = super_search(term, number, page)
    try:
        urls = [
            f['_links']['thumbnail_custom']['href']
            for f in x['_embedded']['mediaTypeResults'][0]['result']['_embedded']['items']
            if f['accessInfo']['accessAllowedFrom'] == 'EVERYWHERE'
               and 'thumbnail_custom' in f['_links']
        ]
    except:
        urls = [' ... hmm ...']
    return urls


def get_picture_from_urn(urn: Union[int, str], width: int = 0, height: int = 300):
    """Fetch the Image object with its URN identifier.

    :meta private:

    :param urn: The uniform resource name number
    :param width: Resolution width of the image.
    :param height: Resolution height of the image.
    :return: An :py:class:`~PIL.Image.Image` object.
    """
    meta = iiif_manifest(urn)
    if 'error' not in meta:
        if width == 0 and height == 0:
            url = f"https://www.nb.no/services/image/resolver/{urn}/full/full/0/native.jpg"
        else:
            url = f"https://www.nb.no/services/image/resolver/{urn}/full/{width},{height}/0/native.jpg"
        # print(url)
    return Image.open(load_picture(url))


def get_picture_from_url(url, width=0, height=300):
    return Image.open(
        load_picture(
            url.format(width=width, height=height)
        )
    )


def get_metadata_from_url(url):
    import re
    urn = re.findall("(URN.*?)(?:/)", url)[0]
    triple = iiif_manifest(urn)
    # print(urn, triple)
    r = dict()
    if not 'error' in triple:
        r = {x['label']: x['value'] for x in triple['metadata'] if 'label' in x}
    else:
        r = triple['error']
    return r


def find_urns(term):
    """From result of super_search, to be fed into iiif_manifest"""

    ss = super_search(term)
    urns = [
        f['metadata']['identifiers']['urn']
        for f in ss['_embedded']['mediaTypeResults'][0]['result']['_embedded']['items']
        if 'urn' in f['metadata']['identifiers']
    ]
    return urns


def total_urls(number=50, page=0):
    """find urls sequentially """
    x = total_search(number, page)
    try:
        urls = [
            f['_links']['thumbnail_custom']['href']
            for f in x['_embedded']['items']
            if f['accessInfo']['accessAllowedFrom'] == 'EVERYWHERE'
               and 'thumbnail_custom' in f['_links']
        ]
    except:
        urls = []
    return urls


def json2html(meta):
    items = ["<dt>{key}</dt><dd>{val}</dd>".format(key=key, val=meta[key]) for key in meta]
    result = "<dl>{items}</dl>".format(items=' '.join(items))
    return result


def display_finds(r):
    """A list of urls in r is displayed as HTML"""
    rows = ["<tr><td><img src='{row}'</td><td>{meta}</td></tr>".format(row=row, meta=json2html(
        get_metadata_from_url(row))).format(width=0, height=200) for row in r]
    return HTML("""<html><head></head>
     <body>
     <table>
     {rows}
     </table>
     </body>
     </html>
     """.format(rows=' '.join(rows)))



import pandas as pd

from dhlab.api.dhlab_api import get_dispersion


class Dispersion:
    """Count occurrences of words in the given URN object."""

    def __init__(
        self, urn: str = None, wordbag: list = None, window: int = 1000, pr: int = 100
    ):
        """Wrapper class for get_dispersion

        :param urn: urn, defaults to None
        :type urn: str, optional
        :param wordbag: words, defaults to None
        :type wordbag: list, optional
        :param window: The number of tokens to search through per row, defaults to 1000
        :type window: int, optional
        :param pr: defaults to 100
        :type pr: int, optional
        """
        if isinstance(wordbag, list):
            dispersion = {
                w: get_dispersion(urn, words=[w], window=window, pr=pr) for w in wordbag
            }
        elif isinstance(wordbag, dict):
            dispersion = {
                w: get_dispersion(urn, words=wordbag[w], window=window, pr=pr)
                for w in wordbag
            }
        elif isinstance(wordbag, str):
            dispersion = {
                wordbag: get_dispersion(urn, words=[wordbag], window=window, pr=pr)
            }
        else:
            dispersion = {}
        self.dispersion = pd.DataFrame(dispersion)

    def plot(self, **kwargs):
        self.dispersion.plot(**kwargs)



"""Tools for querying the Norwegian National Bibliography Marc 21"""
import os
from functools import wraps

import requests

from dhlab.constants import BASE_URL

# TODO: Add support for more fields


def _api_call_deco(service):
    """Decorator for calling a service from DH-lab API

    :param service: Name of service
    """

    def inner_decorator(func):
        """Inner decorator

        :param func: function to decorate. Must return params
        """

        @wraps(func)
        def wrapper(*args, **kwargs):
            params = func(*args, **kwargs)
            return requests.post(os.path.join(BASE_URL, service), json=params).json()

        return wrapper

    return inner_decorator


@_api_call_deco("metadata_query")
def metadata_query(conditions, limit=5):
    """Query the Norwegian National Bibliography using Marc 21 fields and values
    
    Example:
    conditions = [
        ["245", "a", "kongen"],
        ["008", "", "nno"]
        ]

    :param conditions: Marc 21 fields and values to search for
    :type conditions: list of lists
    :param limit: number of records to return, defaults to 5
    :type limit: int, optional
    :return: list of urns
    :rtype: json
    """
    params = {"conditions": conditions, "limit": limit}
    return params


@_api_call_deco("metadata_from_urn")
def metadata_from_urn(urns, fields=None):
    """Gets MARC 21 json for a URN or list of URN

    :param urns: list of URNs
    :param fields: list of marc 21 fields to return
    :return: API call parameters
    """
    params = {"urns": urns, "fields": fields}
    return params


## Utility


def pretty_print_marc21json(record):
    """Prints a record from the Norwegian National Bibliography in a readable format

    :param record: Marc 21 record in json format
    """

    print("Record:")
    for field in record["fields"]:
        for key, val in field.items():
            if "subfields" in val:
                print(key + ":")
                for subfield in val["subfields"]:
                    for subfield_k, subfield_val in subfield.items():
                        print("\t" + subfield_k + ": " + subfield_val)
            else:
                print(key + ": " + val)
    print()



"""Possible future funcs"""
import re

import pandas as pd
from IPython.display import HTML

import dhlab as dh
from dhlab.api.dhlab_api import (
    concordance,
    document_corpus,
    get_document_frequencies,
    get_metadata,
    urn_collocation,
)
from dhlab.text.utils import urnlist


# convert cell to a link
def make_link(row):
    r = "<a target='_blank' href = 'https://urn.nb.no/{x}'>{x}</a>".format(x=str(row))
    return r


# find hits a cell
def find_hits(x):
    return " ".join(re.findall("<b>(.+?)</b", x))


class Corpus(pd.DataFrame):
    @classmethod
    def build(
        self,
        doctype=None,
        author=None,
        freetext=None,
        fulltext=None,
        from_year=None,
        to_year=None,
        from_timestamp=None,
        to_timestamp=None,
        title=None,
        ddk=None,
        subject=None,
        lang=None,
        limit=10,
        order_by="random",
    ):
        res = document_corpus(
            doctype,
            author,
            freetext,
            fulltext,
            from_year,
            to_year,
            from_timestamp,
            to_timestamp,
            title,
            ddk,
            subject,
            lang,
            limit,
            order_by,
        )
        return Corpus(res)

    def extend_from_identifiers(self, identifiers=None):
        new_corpus = get_metadata(urnlist(identifiers))
        return pd.concat([self, new_corpus], axis=0)

    def get_freqs(self):
        return Frequencies.get_freqs(self)

    def get_concordances(self, words, window=20, limit=500):
        return Concordance.get_concordances(self, words, window=window, limit=limit)

    def get_collocations(
        self,
        #  corpus=None,
        words=None,
        before=10,
        after=10,
        reference=None,
        samplesize=20000,
        alpha=False,
        ignore_caps=False,
    ):
        return Collocations.get_collocations(
            corpus=self,
            words=words,
            before=before,
            after=after,
            reference=reference,
            samplesize=samplesize,
            alpha=alpha,
            ignore_caps=ignore_caps,
        )

    @property
    def _constructor(self):
        return Corpus


class Frequencies(pd.DataFrame):
    _metadata = ["_title_dct"]

    @classmethod
    def get_freqs(self, corpus, words=None):
        res = get_document_frequencies(urns=urnlist(corpus), words=words)
        self._title_dct = {k: v for k, v in zip(corpus.dhlabid, corpus.title)}
        return Frequencies(res)

    @property
    def _constructor(self):
        return Frequencies

    def sum_freqs(self):
        return self.sum(axis=1).to_frame("frequencies")

    def display_names(self):
        """Display data with record names as column titles."""
        return self.rename(self._title_dct, axis=1)


class Concordance(pd.DataFrame):
    @property
    def _constructor(self):
        return Concordance

    @classmethod
    def get_concordances(self, corpus, words, window=20, limit=500):
        res = concordance(urns=urnlist(corpus), words=words, window=window, limit=limit)

        res["link"] = res.urn.apply(make_link)

        res.rename({"conc": "concordance"}, axis=1, inplace=True)

        return Concordance(res)

    def show(self, n=10, style=True):
        if style:
            result = self.sample(min(n, len(self)))[["link", "concordance"]].style
        else:
            result = self.sample(min(n, len(self)))
        return result

    def split_view(self, html=False):
        df = self.concordance.str.split("</?b>", expand=True)
        df.rename(
            {0: "left", 1: "hit", 2: "right", 3: "hit2", 4: "right2"},
            axis=1,
            inplace=True,
        )
        df.index = self.urn

        if html:
            return HTML(df.to_html())
        else:
            return df


class Collocations(pd.DataFrame):
    @property
    def _constructor(self):
        return Collocations

    @classmethod
    def get_collocations(
        self,
        corpus=None,
        words=None,
        before=10,
        after=10,
        reference=None,
        samplesize=20000,
        alpha=False,
        ignore_caps=False,
    ):
        if isinstance(words, str):
            words = [words]

        res = pd.concat(
            [
                urn_collocation(
                    urns=urnlist(corpus),
                    word=w,
                    before=before,
                    after=after,
                    samplesize=samplesize,
                )
                for w in words
            ]
        )[["counts"]]

        if reference is not None:
            teller = res.counts / res.counts.sum()
            divisor = reference.iloc[:, 0] / reference.iloc[:, 0].sum()
            res["relevance"] = teller / divisor

        return Collocations(res)

    def show(self, sortby="counts", n=20):
        return self.sort_values(by=sortby, ascending=False).head(n)



from typing import List, Union

import pandas as pd
from pandas import DataFrame

# import dhlab as dh
# from dhlab.text.conc_coll import Concordance, Collocations, Counts
import dhlab.text.conc_coll as dh
from dhlab.api.dhlab_api import document_corpus, evaluate_documents, get_metadata
from dhlab.text.dhlab_object import DhlabObj
from dhlab.text.utils import urnlist


class Corpus(DhlabObj):
    """Class representing as DHLAB Corpus

    Primary object for working with dhlab data. Contains references to texts
    in National Library's collections and metadata about them.
    Use with `.coll`, `.conc` or `.freq` to analyse using dhlab tools.
    """

    doctypes = [
        "digibok",
        "digavis",
        "digitidsskrift",
        "digistorting",
        "digimanus",
        "kudos",
    ]

    def __init__(
        self,
        doctype=None,
        author=None,
        freetext=None,
        fulltext=None,
        from_year=None,
        to_year=None,
        from_timestamp=None,
        to_timestamp=None,
        title=None,
        ddk=None,
        subject=None,
        lang=None,
        limit=10,
        order_by="random",
    ):
        """Create Corpus

        :param str doctype: ``"digibok"``, ``"digavis"``, \
            ``"digitidsskrift"`` or ``"digistorting"``
        :param str author: Name of an author.
        :param str freetext: any of the parameters, for example:\
            ``"digibok AND Ibsen"``.
        :param str fulltext: words within the publication.
        :param int from_year: Start year for time period of interest.
        :param int to_year: End year for time period of interest.
        :param int from_timestamp: Start date for time period of interest.
            Format: ``YYYYMMDD``, books have ``YYYY0101``
        :param int to_timestamp: End date for time period of interest.
            Format: ``YYYYMMDD``, books have ``YYYY0101``
        :param str title: Name or title of a document.
        :param str ddk: `Dewey Decimal Classification \
            <https://no.wikipedia.org/wiki/Deweys_desimalklassifikasjon>`\
                _ identifier.
        :param str subject: subject (keywords) of the publication.
        :param str lang: Language of the publication, as a 3-letter ISO code.
            Example: ``"nob"`` or ``"nno"``
        :param int limit: number of items to sample.
        """

        if (
            doctype
            or author
            or freetext
            or fulltext
            or from_year
            or to_year
            or from_timestamp
            or to_timestamp
            or title
            or ddk
            or lang
        ):
            self.corpus = document_corpus(
                doctype,
                author,
                freetext,
                fulltext,
                from_year,
                to_year,
                from_timestamp,
                to_timestamp,
                title,
                ddk,
                subject,
                lang,
                limit,
                order_by,
            )

        else:
            self.corpus = pd.DataFrame(columns=["urn"])

        super().__init__(self.corpus)
        # self.size = len(self.corpus)

        self.frame.rename(
            columns={
                "urn": "urn",
                "authors": "author",
                "langs": "language",
                "genres": "genre",
            },
            inplace=True,
        )

    @classmethod
    def from_identifiers(cls, identifiers: List[Union[str, int]]):
        """Construct Corpus from list of identifiers"""
        corpus = Corpus()
        corpus.extend_from_identifiers(identifiers=identifiers)
        return corpus

    @classmethod
    def from_df(cls, df: DataFrame, check_for_urn: bool = False):
        """Typecast Pandas DataFrame to Corpus class

        DataFrame most contain URN column"""

        # If Series, return as is
        if isinstance(df, pd.Series):
            return df

        df = df.copy()  # Avoid modifying original DataFrame

        corpus = Corpus()
        if check_for_urn:
            corpus.corpus = cls._urn_id_in_dataframe_cols(df)
        else:
            corpus.corpus = df
        corpus.frame = corpus.corpus
        return corpus

    @classmethod
    def from_csv(cls, path: str):
        """Import corpus from csv"""
        df = pd.read_csv(path)
        return cls.from_df(df)

    @staticmethod
    def _urn_id_in_dataframe_cols(dataframe: Union[DataFrame, type("Corpus")]) -> DataFrame:
        """Checks if dataframe contains URN column"""
        if "urn" in dataframe.columns:
            if dataframe.urn.str.contains("^URN:NBN:no-nb_.+").all():
                return dataframe
        raise ValueError("No'urn'-column in dataframe.")

    def extend_from_identifiers(self, identifiers: list = None):
        new_corpus = get_metadata(urnlist(identifiers))
        self.add(new_corpus)

    def evaluate_words(self, wordbags=None):
        df = evaluate_documents(wordbags=wordbags, urns=list(self.corpus.urn))
        df.index = df.index.astype(int)
        cols = df.columns
        df = pd.concat(
            [df, self.corpus[["dhlabid", "urn"]].set_index("dhlabid")], axis=1
        )
        df = df.set_index("urn")
        return df[cols].fillna(0)

    def add(self, new_corpus: Union[DataFrame, type("Corpus")]):
        """Utility for appending Corpus or DataFrame to self"""
        if self._is_Corpus(new_corpus):
            new_corpus = new_corpus.frame
        self.frame = (
            pd.concat([self.frame, new_corpus]).drop_duplicates().reset_index(drop=True)
        )
        self.corpus = self.frame
        # self.size = len(self.frame)

    def sample(self, n: int = 5):
        """Create random subkorpus with `n` entries"""
        n = min(n, self.size)
        sample = self.corpus.sample(n).copy()
        return self.from_df(sample)

    def only_one_author(self):
        """Only select items with one author"""
        mask = self.frame.author.apply(lambda x: len(x.split('/'))) == 1
        return self.from_df(self.frame[mask])

    def conc(self, words, window: int = 20, limit: int = 500) -> dh.Concordance:
        """Get concodances of `words` in corpus"""
        return dh.Concordance(
            corpus=self.frame, query=words, window=window, limit=limit
        )

    def coll(
        self,
        words=None,
        before=10,
        after=10,
        reference=None,
        samplesize=20000,
        alpha=False,
        ignore_caps=False,
    ) -> dh.Collocations:
        """Get collocations of `words` in corpus"""
        return dh.Collocations(
            corpus=self.frame,
            words=words,
            before=before,
            after=after,
            reference=reference,
            samplesize=samplesize,
            alpha=alpha,
            ignore_caps=ignore_caps,
        )

    def count(self, words=None):
        """Get word frequencies for corpus"""
        return dh.Counts(self, words)

    def freq(self, words=None):
        """Get word frequencies for corpus"""
        return dh.Counts(self, words)

    @staticmethod
    def _is_Corpus(corpus: "Corpus") -> bool:
        """Check if `input` is Corpus or DataFrame"""
        if type(corpus) not in [DataFrame, Corpus]:
            raise TypeError("Input is not Corpus or DataFrame")
        return isinstance(corpus, Corpus)

    def __add__(self, other):
        """Add two Corpus objects"""
        if not self._is_Corpus(other):
            raise TypeError("Input is not Corpus or DataFrame")
        return self.from_df(
            pd.concat([self.frame, other.frame])
            .drop_duplicates()
            .reset_index(drop=True)
        )

    def _make_subcorpus(self, **kwargs) -> "Corpus":
        dct = kwargs.copy()
        year_range = dct.pop("year_range", None)

        for key in dct.keys():
            if key not in self.frame.columns:
                print(f"Key {key} not in corpus")
                return None

        # Make result dataframe
        res = self.frame.copy()

        # Get year range
        if year_range is not None:
            y1 = int(year_range[0])
            y2 = int(year_range[1])

            # Apply year range
            res = res.loc[res["year"] >= y1].loc[res["year"] <= y2]

        for key, val in dct.items():
            res = res.loc[res[key].str.contains(val)]

        return self.from_df(res)

    def make_subcorpus(self, authors: str = None, title: str = None) -> "Corpus":
        """Make subcorpus based on author and title

        Args:
            authors (str, optional): search for author field. Defaults to None.
            title (str, optional): search title field. Defaults to None.

        Returns:
            Corpus: A subset of the original corpus
        """
        dct = {}
        if authors is not None:
            dct["author"] = authors
        if title is not None:
            dct["title"] = title

        return self._make_subcorpus(**dct)

    def check_integrity(self):
        """Check the integrity of the corpus data."""

        def test_dhlabid_series(series: pd.Series) -> bool:
            """Check if dhlabid series is valid"""
            if not series.apply(lambda x: isinstance(x, int)).all():
                return False
            if not ((series >= 1e8) & (series < 1e9)).all():
                return False

            return True

        def test_urn_series(series: pd.Series) -> bool:
            """Check if URN series is valid"""
            if series.str.startswith("URN:NBN:no-nb_").all():
                return True
            try:
                series = series.apply(lambda x: int(x))
                return test_dhlabid_series(series)
            except:
                return False

        # Check if the DataFrame is empty
        if self.corpus.empty:
            raise ValueError("Corpus is empty.")

        # Check for the presence of essential columns
        required_columns = ["urn", "dhlabid", "author", "language", "genre"]
        for col in required_columns:
            if col not in self.corpus.columns:
                raise ValueError(f"Essential column '{col}' is missing.")

        # Validate dhlabid format
        if not test_dhlabid_series(self.corpus["dhlabid"]):
            raise ValueError("Some dhlabid values are in an incorrect format.")

        # Validate URN format
        if not test_urn_series(self.corpus["urn"]):
            raise ValueError("Some URN values are in an incorrect format.")

        return True



"""Viz tools for for pandas DataFrame"""
import pandas as pd
import seaborn as sns


def heatmap(df: pd.DataFrame, color: str = "green"):
    """A wrapper for heatmap of a dataframe `df`."""
    return df.fillna(0).style.background_gradient(
        cmap=sns.light_palette(color, as_cmap=True)
    )



# api
from dhlab.api.dhlab_api import totals

# legacy code
from dhlab.legacy import (
    graph_networkx_louvain,
    module_update,
    nb_external_files,
    nbpictures,
    nbtext,
    token_map,
)

# metadata
from dhlab.metadata.natbib import metadata_from_urn, metadata_query

# ngram
from dhlab.ngram.ngram import Ngram, NgramBook, NgramNews

# text
from dhlab.text import nbtokenizer
from dhlab.text.chunking import Chunks
from dhlab.text.conc_coll import Collocations, Concordance, Counts
from dhlab.text.corpus import Corpus
from dhlab.text.geo_data import GeoData, GeoNames
from dhlab.text.parse import NER, POS, Models
from dhlab.text.wildcard import WildcardWordSearch
from dhlab.utils.display import css
from dhlab.utils.files import download_from_github, get_file_from_github

# wordbank
from dhlab.wordbank.wordbank import WordForm, WordLemma, WordParadigm



import json
from collections import Counter

import matplotlib.pyplot as plt

try:
    from wordcloud import WordCloud
except BaseException:
    print("wordcloud er ikke installert, kan ikke lage ordskyer")


def make_cloud(json_text, top=100, background='white',
               stretch=lambda x: 2 ** (10 * x), width=500, height=500,
               font_path=None):
    """Create a word cloud from a frequency list."""
    pairs0 = Counter(json_text).most_common(top)
    pairs = {x[0]: stretch(x[1]) for x in pairs0}
    wc = WordCloud(
        font_path=font_path,
        background_color=background,
        width=width,
        # color_func=my_colorfunc,
        ranks_only=True,
        height=height).generate_from_frequencies(pairs)
    return wc


def draw_cloud(sky, width=20, height=20, fil=''):
    """Draw a word cloud produced by :func:`make_cloud`."""
    plt.figure(figsize=(width, height))
    plt.imshow(sky, interpolation='bilinear')
    figplot = plt.gcf()
    if fil != '':
        figplot.savefig(fil, format='png')


def cloud(df, column='', top=200, width=1000, height=1000, background='black',
          file='', stretch=10, font_path=None):
    """Make and draw a cloud from a pandas dataframe, using :func:`make_cloud` and
    :func:`draw_cloud`."""
    if column == '':
        column = df.columns[0]
    data = json.loads(df[column].to_json())
    a_cloud = make_cloud(data, top=top,
                         background=background, font_path=font_path,
                         stretch=lambda x: 2 ** (stretch * x), width=width,
                         height=height)
    draw_cloud(a_cloud, fil=file)


import pandas as pd

from dhlab.api.dhlab_api import ner_from_urn, pos_from_urn, show_spacy_models


class Models:
    """Show the spaCy language models available"""
    def __init__(self):
        self.models = show_spacy_models()
        
class NER:
    """Provide NER """
    def __init__(self, urn=None, model=None, start_page=0, to_page=0):
        self.model = model
        self.ner = ner_from_urn(urn=urn, model=self.model, start_page=start_page, to_page=to_page)
        
class POS:
    """Provide POS and a parse"""
    def __init__(self, urn=None, model=None, start_page=0, to_page=0):
        self.model = model
        self.pos = pos_from_urn(urn=urn, model=self.model, start_page=start_page, to_page=to_page)




import os
from urllib.parse import urlparse

import requests
from IPython.display import HTML, display, Markdown


def code_toggle(button_text="Klikk for å vise/skjule kodeceller"):
    display(
        HTML(
            '''<div>
                <style>
                 .mybutton {
                    background-color: lightgrey;
                    border: none;
                    color: white;
                    padding: 10px 16px;
                    text-align: center;
                    text-decoration: none;
                    display: inline-block;
                    font-size: 16px;
                    margin: 4px 2px;
                    cursor: pointer;
                }
            </style>
            <script>
                code_show=true;
                function code_toggle() {
                 if (code_show){
                 $('div.input').hide();
                 } else {
                 $('div.input').show();
                 }
                 code_show = !code_show
                }
                $( document ).ready(code_toggle);
            </script>
            <form  action="javascript:code_toggle()">
                <input class='mybutton' type="submit" value=''' + '"' + button_text + '"' + '''>
            </form>
        </div>'''
        ))


def printmd(S):
    display(Markdown(S))


def update(module="", overwrite=True, silent=False):
    """Fetch modules from Github and write them to folder"""
    nba = requests.get(
        f"https://raw.githubusercontent.com/Yoonsen/Modules/master/{module}.py",
        headers={'Cache-Control': 'no-cache'}
    )
    filename = f'{module}.py'
    if nba.status_code == 200:
        file_exists = os.path.exists(filename)
        if file_exists and not overwrite:
            if not silent:
                printmd(
                    ("File {f} exists - call `update('{m}', overwrite = True)`"
                     " in order to download module `{m}` anyway").format(
                        f=os.path.abspath(filename), m=module))
        else:
            nba = nba.text
            with open(filename, 'w', encoding='UTF-8') as pyfile:
                pyfile.write(nba)
                pyfile.flush()
                pyfile.close()
            if not silent:
                printmd(f"Updated file `{os.path.abspath(module)}.py`")
    else:
        printmd(
            f"An error occured during download for {module} with error "
            f"{nba.status_code}"
        )


def css(url=None):
    """Associate a css stylesheet with the notebook.

    Specify a file or web reference, default is a custom css.
    """
    url = ("https://raw.githubusercontent.com/Yoonsen/Modules/master"
           "/css_style_sheets/nb_notebook.css") if url is None else url
    uri = urlparse(url)
    css_file = ""

    if uri.scheme.startswith('http'):
        query = requests.get(url)
        if query.status_code == 200:
            css_file = query.text

    elif uri.scheme == "file":
        # assume on form "file:/// on windows there is drive letter on unix
        # not"
        file_path = url[7:]
        if file_path[2] == ':':  # then windows drive reference
            file_path = file_path[1:]
        with open(file_path, encoding='utf-8') as file:
            css_file = file.read()
    else:
        # assume string is a file locator
        with open(url, encoding='utf-8') as file:
            css_file = file.read()

    return HTML(f"<style>{css_file}</style>")


if __name__ == '__main__':
    update("nbtext", overwrite=False, silent=True)
    update("graph_networkx_louvain", overwrite=False, silent=True)
    update('token_map', overwrite=False, silent=True)



"""Text analysis functionality, including building text corpora."""
from .chunking import Chunks
from .conc_coll import Collocations, Concordance, Counts
from .corpus import Corpus
from .dispersion import Dispersion
from .geo_data import GeoData
from .nbtokenizer import Tokens, tokenize
from .parse import NER, POS, Models
from .wildcard import WildcardWordSearch



from datetime import datetime

from dhlab.api.dhlab_api import ngram_book, ngram_news
from dhlab.ngram.nb_ngram import nb_ngram
from dhlab.text.dhlab_object import DhlabObj


class Ngram(DhlabObj):
    """Top level class for ngrams"""

    def __init__(
        self,
        words=None,
        from_year=None,
        to_year=None,
        doctype="bok",
        mode="relative",
        lang="nob",
        **kwargs
    ):
        """Ngram builder class.

        Build Ngrams from the National Librarys collections.
        Use with book corpus or newspaper corpus.
        Lang parameter is only supported for book (`bok`) corpus.
        Defaults to `None` if doctype is `avis`.

        :param words: words to examine, defaults to None
        :type words: str or list of str, optional
        :param from_year: lower period cutoff, defaults to None
        :type from_year: int, optional
        :param to_year: upper period cutoff, defaults to None
        :type to_year: int, optional
        :param doctype: `bok` or `avis` , defaults to 'bok'
        :type doctype: str, optional
        :param mode: Frequency measure, defaults to 'relative'
        :type mode: str, optional
        :param lang: `nob`, `nno`. Only use with docytype='bok', defaults to 'nob'
        :type lang: str, optional
        :param kwargs: Keyword arguments for  Ngram._ipython_display_() Ngram.plot()
        """

        self.date = datetime.now()
        if to_year is None:
            to_year = self.date.year
        if from_year is None:
            from_year = 1950

        self.from_year = from_year
        self.to_year = to_year
        self.words = words
        self.lang = lang
        if doctype is not None:
            if "bok" in doctype:
                doctype = "bok"
            elif "avis" in doctype:
                doctype = "avis"
            else:
                doctype = "bok"
        else:
            doctype = "bok"

        # Set default lang for 'bok'-corpus
        if doctype == "avis":
            lang = None

        ngrm = nb_ngram(
            terms=", ".join(words),
            corpus=doctype,
            years=(from_year, to_year),
            smooth=1,
            lang=lang,
            mode=mode,
        )
        ngrm.index = ngrm.index.astype(str)
        self.ngram = ngrm

        self.kwargs = kwargs

        super().__init__(self.ngram)

    def plot(self, smooth=4, **kwargs):
        """:param smooth: smoothing the curve"""
        grf = self.ngram.rolling(window=smooth, win_type="triang").mean()
        grf.plot(**kwargs)

    def compare(self, another_ngram):
        """Divide one ngram by another - measures difference"""
        start_year = max(
            datetime(self.from_year, 1, 1), datetime(another_ngram.from_year, 1, 1)
        ).year
        end_year = min(
            datetime(self.to_year, 1, 1), datetime(another_ngram.to_year, 1, 1)
        ).year
        transposed_ngram = self.ngram.loc[str(start_year) : str(end_year)].transpose()
        sum_other_ngram = (
            another_ngram.ngram[str(start_year) : str(end_year)].transpose().sum()
        )
        compare = (transposed_ngram / sum_other_ngram).transpose()
        return compare

    def _ipython_display_(self):
        self.plot(**self.kwargs)


class NgramBook(Ngram):
    """Extract ngrams using metadata with functions to be inherited."""

    def __init__(
        self,
        words=None,
        title=None,
        publisher=None,
        city=None,
        lang="nob",
        from_year=None,
        to_year=None,
        ddk=None,
        subject=None,
        **kwargs
    ):
        """Create Dhlab Ngram from metadata

        :param words: words to examine, defaults to None
        :type words: str or list of str optional
        :param title: _description_, defaults to None
        :type title: _type_, optional
        :param publisher: _description_, defaults to None
        :type publisher: _type_, optional
        :param city: _description_, defaults to None
        :type city: _type_, optional
        :param lang: _description_, defaults to 'nob'
        :type lang: str, optional
        :param from_year: _description_, defaults to None
        :type from_year: _type_, optional
        :param to_year: _description_, defaults to None
        :type to_year: _type_, optional
        :param ddk: _description_, defaults to None
        :type ddk: _type_, optional
        :param subject: _description_, defaults to None
        :type subject: _type_, optional
        :return: _description_
        :rtype: _type_
        """

        super().__init__(
            words,
            from_year=from_year,
            to_year=to_year,
            lang=lang,
            doctype="bok",
            **kwargs
        )
        self.date = datetime.now()
        if to_year is None:
            to_year = self.date.year
        if from_year is None:
            from_year = 1950
        self.from_year = from_year
        self.to_year = to_year
        self.words = words
        self.title = title
        self.publisher = publisher
        self.city = city
        self.lang = lang
        self.ddk = ddk
        self.subject = subject
        self.ngram = ngram_book(
            word=words,
            title=title,
            publisher=publisher,
            lang=lang,
            city=city,
            period=(from_year, to_year),
            ddk=ddk,
            topic=subject,
        )
        # update frame attribute
        self.frame = self.ngram
        # self.cohort =  (self.ngram.transpose()/self.ngram.transpose().sum()).transpose()


class NgramNews(Ngram):
    def __init__(
        self, words=None, title=None, city=None, from_year=None, to_year=None, **kwargs
    ):
        super().__init__(
            words, from_year=from_year, to_year=to_year, doctype="avis", **kwargs
        )
        self.date = datetime.now()
        self.from_year = 1950 if from_year is None else from_year
        self.to_year = self.date.year if to_year is None else to_year
        self.words = words
        self.title = title
        self.ngram = ngram_news(word=words, title=title, period=(from_year, to_year))
        # update frame attribute
        self.frame = self.ngram
        # self.cohort =  (self.ngram.transpose()/self.ngram.transpose().sum()).transpose()



from io import StringIO
from typing import Dict, List, Tuple, Union

import pandas as pd
import requests

# from requests import HTTPError, JSONDecodeError, ConnectionError
from pandas import DataFrame, Series

from dhlab.constants import BASE_URL

pd.options.display.max_rows = 100

# wildcard search for words


def wildcard_search(word, factor=2, freq_limit=10, limit = 50):
    res = requests.get(f"{BASE_URL}/wildcard_word_search", 
                        params={
                            'word':word, 
                            'factor':factor, 
                            'freq_lim':freq_limit, 
                            'limit':limit
                        }
                       )
    #columns = ["key", "name", "alternatename", "latitude", "longitude", "feature class", "feature code"]
    return pd.DataFrame.from_dict(res.json(), orient = 'index', columns=['freq'])





# fetch metadata

def images(text = None, part=True):
    """ Retrive images from bokhylla
    :param text: fulltext query expression for sqlite
    :param part: if a number the whole page is shown
    ... bug prevents these from going thru
    :param delta: if part=True then show additional pixels around image
    :parsm hits: number of images"""

    params = locals()
    r = requests.get(f"{BASE_URL}/images", params=params)
    js = r.json()
    return js

def ner_from_urn(urn: str = None, model: str = None, start_page = 0, to_page = 0) -> DataFrame:
    """Get NER annotations for a text (``urn``) using a spacy ``model``.

    :param str urn: uniform resource name, example: ``URN:NBN:no-nb_digibok_2011051112001``
    :param str model: name of a spacy model.
        Check which models are available with :func:`show_spacy_models`
    :return: Dataframe with annotations and their frequencies
    """

    params = locals()
    r = requests.get(f"{BASE_URL}/ner_urn", params=params)
    df = pd.read_json(r.json())
    return df


def pos_from_urn(urn: str = None, model: str = None, start_page = 0, to_page = 0) -> DataFrame:
    """Get part of speech tags and dependency parse annotations for a text (``urn``) with a SpaCy ``model``.

    :param str urn: uniform resource name, example: ``URN:NBN:no-nb_digibok_2011051112001``
    :param str model: name of a spacy model.
        Check which models are available with :func:`show_spacy_models`
    :return: Dataframe with annotations and their frequencies
    """
    params = locals()
    r = requests.get(f"{BASE_URL}/pos_urn", params=params)
    df = pd.read_json(r.json())
    return df


def show_spacy_models() -> List:
    """Show available SpaCy model names."""
    try:
        r = requests.get(f"{BASE_URL}/ner_models")
        #r.raise_for_status()
        res = r.json()
    except: #(HTTPError, JSONDecodeError, ConnectionError) as error:
        #print(error.__doc__, error)
        print("Server-request gikk ikke gjennom. Kan ikke vise SpaCy-modellnavn.")
        res =  []
    return res

def get_places(urn: str) -> DataFrame:
    """Look up placenames in a specific URN.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/places <https://api.nb.no/dhlab/#/default/post_places>`_.

    :param str urn: uniform resource name, example: ``URN:NBN:no-nb_digibok_2011051112001``
    """
    params = locals()
    r = requests.post(f"{BASE_URL}/places", json=params)
    # print(r.status_code)
    return pd.DataFrame(r.json())


def geo_lookup(
        places: List,
        feature_class: str = None,
        feature_code: str = None,
        field: str = "alternatename",
) -> DataFrame:
    """From a list of places, return their geolocations

    :param list places: a list of place names - max 1000
    :param str feature_class: which GeoNames feature class to return. Example: ``P``
    :param str feature_code: which GeoNames feature code to return. Example: ``PPL``
    :param str field: which name field to match - default "alternatename".
    """
    res = requests.post(
        f"{BASE_URL}/geo_data",
        json={
            "words": places,
            "feature_class": feature_class,
            "feature_code": feature_code,
            "field": field,
        },
    )
    columns = [
        "geonameid",
        "name",
        "alternatename",
        "latitude",
        "longitude",
        "feature_class",
        "feature_code",
    ]
    return pd.DataFrame(res.json(), columns=columns)


def get_dispersion(
        urn: str = None,
        words: List = None,
        window: int = 300,
        pr: int = 100,
) -> Series:
    """Count occurrences of words in the given URN object.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint ``/dispersion``.

    :param str urn: uniform resource name, example: ``URN:NBN:no-nb_digibok_2011051112001``
    :param list words: list of words. Defaults to a list of punctuation marks.
    :param int window: The number of tokens to search through per row. Defaults to 300.
    :param int pr: defaults to 100.
    :return: a ``pandas.Series`` with frequency counts of the words in the URN object.
    """
    params = locals()
    r = requests.post(f"{BASE_URL}/dispersion", json=params)
    return pd.Series(r.json())


def get_metadata(urns: List[str] = None) -> DataFrame:
    """Get metadata for a list of URNs.

    Calls the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/get_metadata <https://api.nb.no/dhlab/#/default/post_get_metadata>`_.

    :param list urns: list of uniform resource name strings, for example:
        ``["URN:NBN:no-nb_digibok_2008051404065", "URN:NBN:no-nb_digibok_2010092120011"]``
    """
    params = locals()
    r = requests.post(f"{BASE_URL}/get_metadata", json=params)
    return DataFrame(r.json())

def get_identifiers(identifiers: list = None) -> list:
    """Convert a list of identifiers, oaiid, sesamid, urns or isbn10 to dhlabids"""
    res = requests.post(f"{BASE_URL}/identifiers", json={'identifiers':[i for i in identifiers if i != '']})
    return res.json()

def get_chunks(urn: str = None, chunk_size: int = 300) -> Union[Dict, List]:
    """Get the text in the document ``urn`` as frequencies of chunks
     of the given ``chunk_size``.

    Calls the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    ``/chunks``.

    :param str urn: uniform resource name, example: ``URN:NBN:no-nb_digibok_2011051112001``
    :param int chunk_size: Number of tokens to include in each chunk.
    :return: list of dicts with the resulting chunk frequencies, or an empty dict
    """

    if urn is None:
        return {}
    r = requests.get(f"{BASE_URL}/chunks", params=locals())
    if r.status_code == 200:
        result = r.json()
    else:
        result = {}
    return result


def get_chunks_para(urn: str = None) -> Union[Dict, List]:
    """Fetch chunks and their frequencies from paragraphs in a document (``urn``).

    Calls the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    ``/chunks_para``.

    :param str urn: uniform resource name, example: ``URN:NBN:no-nb_digibok_2011051112001``
    :return: list of dicts with the resulting chunk frequencies, or an empty dict
    """

    if urn is None:
        return {}
    r = requests.get(f"{BASE_URL}/chunks_para", params=locals())
    if r.status_code == 200:
        result = r.json()
    else:
        result = {}
    return result


def evaluate_documents(wordbags: Dict = None, urns: List[str] = None) -> DataFrame:
    """Count and aggregate occurrences of topic ``wordbags`` for each document in a list of ``urns``.

    :param dict wordbags: a dictionary of topic keywords and lists of associated words.
        Example: ``{"natur": ["planter", "skog", "fjell", "fjord"], ... }``
    :param list urns: uniform resource names, for example:
        ``["URN:NBN:no-nb_digibok_2008051404065", "URN:NBN:no-nb_digibok_2010092120011"]``
    :return: a ``pandas.DataFrame`` with the topics as columns, indexed by the dhlabids of the
        documents.
    """
    res = requests.post(
        f"{BASE_URL}/evaluate", json={"wordbags": wordbags, "urns": urns}
    )
    df = pd.DataFrame(res.json()).transpose()
    return df


def get_reference(
        corpus: str = "digavis",
        from_year: int = 1950,
        to_year: int = 1955,
        lang: str = "nob",
        limit: int = 100000,
) -> DataFrame:
    """Reference frequency list of the n most frequent words from a given corpus in a given period.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/reference_corpus <https://api.nb.no/dhlab/#/default/get_reference_corpus>`_.

    :param str corpus: Document type to include in the corpus, can be either ``'digibok'`` or
        ``'digavis'``.
    :param int from_year: Starting point for time period of the corpus.
    :param int to_year: Last year of the time period of the corpus.
    :param str lang: Language of the corpus, can be one of
        ``'nob,', 'nno,', 'sme,', 'sma,', 'smj', 'fkv'``
    :param int limit: Maximum number of most frequent words.
    :return: A ``pandas.DataFrame`` with the results.
    """
    params = locals()
    r = requests.get(BASE_URL + "/reference_corpus", params=params)
    if r.status_code == 200:
        result = r.json()
    else:
        result = []
    return pd.DataFrame(result, columns=["word", "freq"]).set_index("word")


def find_urns(docids: Union[Dict, DataFrame] = None, mode: str = "json") -> DataFrame:
    """Return a list of URNs from a collection of docids.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/find_urn`.

    :param docids: dictionary of document IDs (``{docid: URN}``) or a ``pandas.DataFrame``.
    :param str mode: Default 'json'.
    :return: the URNs that were found, in a ``pandas.DataFrame``.
    """
    params = locals()
    r = requests.post(BASE_URL + "/find_urn", json=params)
    if r.status_code == 200:
        res = pd.DataFrame.from_dict(r.json(), orient="index", columns=["urn"])
    else:
        res = pd.DataFrame()
    return res


def _ngram_doc(
        doctype: str = None,
        word: Union[List, str] = ["."],
        title: str = None,
        period: Tuple[int, int] = None,
        publisher: str = None,
        lang: str = None,
        city: str = None,
        ddk: str = None,
        topic: str = None,
) -> DataFrame:
    """Count occurrences of one or more words over a time period.

    The type of document to search through is decided by the ``doctype``.
    Filter the selection of documents with metadata.
    Use % as wildcard where appropriate - no wildcards in ``word`` or ``lang``.

    :param str doctype: API endpoint for the document type to get ngrams for.
        Can be ``'book'``, ``'periodicals'``, or ``'newspapers'``.
    :param word: Word(s) to search for.
        Can be several words in a single string, separated by comma, e.g. ``"ord,ordene,orda"``.
    :type word: str or list of str
    :param str title: Title of a specific document to search through.
    :param period: Start and end years or dates of a time period,
        given as ``(YYYY, YYYY)`` or ``(YYYYMMDD, YYYYMMDD)``.
    :type period: tuple of ints
    :param str publisher: Name of a publisher.
    :param str lang: Language as a 3-letter ISO code (e.g. ``"nob"`` or ``"nno"``)
    :param str city: City of publication.
    :param str ddk: `Dewey Decimal Classification
        <https://no.wikipedia.org/wiki/Deweys_desimalklassifikasjon>`_ identifier.
    :param str topic: Topic of the documents.
    :return: a ``pandas.DataFrame`` with the resulting frequency counts of the word(s),
        spread across years. One year per row.
    """
    params = locals()
    if isinstance(word, str):
        # assume a comma separated string
        word = [w.strip() for w in word.split(",")]
    params["word"] = tuple(word)
    params = {x: params[x] for x in params if not params[x] is None}
    r = requests.post(BASE_URL + "/ngram_" + doctype, json=params)
    # print(r.status_code)
    df = pd.DataFrame.from_dict(r.json(), orient="index")
    df.index = df.index.map(lambda x: tuple(x.split()))
    columns = df.index.levels[0]
    df = pd.concat([df.loc[x] for x in columns], axis=1)
    df.columns = columns
    # df.index = df.index.map(pd.Timestamp)
    return df


def reference_words(
        words: List = None,
        doctype: str = "digibok",
        from_year: Union[str, int] = 1800,
        to_year: Union[str, int] = 2000,
) -> DataFrame:
    """Collect reference data for a list of words over a time period.

    Reference data are the absolute and relative frequencies of the ``words``
    across all documents of the given ``doctype`` in the given time period
    (``from_year`` - ``to_year``).

    :param list words: list of word strings.
    :param str doctype: type of reference document. Can be ``"digibok"`` or ``"digavis"``.
        Defaults to ``"digibok"``.

        .. note::
           If any other string is given as the ``doctype``,
           the resulting data is equivalent to what you get with
           ``doctype="digavis"``.

    :param int from_year: first year of publication
    :param int to_year: last year of publication
    :return: a DataFrame with the words' frequency data
    """
    params = locals()
    r = requests.post(f"{BASE_URL}/reference_words", json=params)
    print(r.status_code, BASE_URL)
    if r.status_code == 200:
        res = pd.DataFrame(r.json(), columns=["word", "freq", "relative"])
    else:
        res = pd.DataFrame()
    return res


def _ngram_doc(
        doctype: str = None,
        word: Union[List, str] = ['.'],
        title: str = None,
        period: Tuple[int, int] = None,
        publisher: str = None,
        lang: str = None,
        city: str = None,
        ddk: str = None,
        topic: str = None
) -> pd.DataFrame:
    """Count occurrences of one or more words over a time period.

    The type of document to search through is decided by the ``endpoint``.
    Filter the selection of documents with metadata.
    Use % as wildcard where appropriate - no wildcards in ``word`` or ``lang``.

    :param str doctype: API endpoint for the document type to get ngrams for.
        Can be ``'book'``, ``'periodicals'``, or ``'newspapers'``.
    :param word: Word(s) to search for.
        Can be several words in a single string, separated by comma.
    :type word: str or list of str
    :param title: Title of a specific document to search through.
    :param tuple[int,int] period: Start and end years of a time period,
        given as ``(start year, end year)``.
    :param str publisher: Name of a publisher.
    :param str lang: Language as a 3-letter ISO code (e.g. ``"nob"`` or ``"nno"``)
    :param str city: City of publication.
    :param str ddk: `Dewey Decimal Classification
        <https://no.wikipedia.org/wiki/Deweys_desimalklassifikasjon>`_ identifier.
    :param str topic: Topic of the documents.
    :return: a `pandas.DataFrame` with the resulting frequency counts of the word(s),
        spread across years. One year per row.
    """
    params = locals()
    if isinstance(word, str):
        # assume a comma separated string
        word = [w.strip() for w in word.split(',')]
    params['word'] = tuple(word)
    params = {x: params[x] for x in params if not params[x] is None}
    r = requests.post(BASE_URL + "/ngram_" + doctype, json=params)
    # print(r.status_code)
    df = pd.DataFrame.from_dict(r.json(), orient='index')
    df.index = df.index.map(lambda x: tuple(x.split()))
    columns = df.index.levels[0]
    df = pd.concat([df.loc[x] for x in columns], axis=1)
    df.columns = columns
    # df.index = df.index.map(pd.Timestamp)
    return df


# @_docstring_parameters_from(_ngram_doc, drop="doctype")
def ngram_book(
        word: Union[List, str] = ["."],
        title: str = None,
        period: Tuple[int, int] = None,
        publisher: str = None,
        lang: str = None,
        city: str = None,
        ddk: str = None,
        topic: str = None,
) -> DataFrame:
    """Count occurrences of one or more words in books over a given time period.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/ngram_book`.

    Filter the selection of books with metadata.
    Use % as wildcard where appropriate - no wildcards in ``word`` or ``lang``.

    :param word: Word(s) to search for.
        Can be several words in a single string, separated by comma, e.g. ``"ord,ordene,orda"``.
    :type word: str or list of str
    :param str title: Title of a specific document to search through.
    :param period: Start and end years or dates of a time period,
        given as ``(YYYY, YYYY)`` or ``(YYYYMMDD, YYYYMMDD)``.
    :type period: tuple of ints
    :param str publisher: Name of a publisher.
    :param str lang: Language as a 3-letter ISO code (e.g. ``"nob"`` or ``"nno"``)
    :param str city: City of publication.
    :param str ddk: `Dewey Decimal Classification
        <https://no.wikipedia.org/wiki/Deweys_desimalklassifikasjon>`_ identifier.
    :param str topic: Topic of the documents.
    :return: a ``pandas.DataFrame`` with the resulting frequency counts of the word(s),
        spread across years. One year per row.
    """
    params = locals()
    if isinstance(word, str):
        # assume a comma separated string
        word = [w.strip() for w in word.split(',')]
    params['word'] = tuple(word)
    params = {x: params[x] for x in params if not params[x] is None}
    r = requests.post(BASE_URL + "/ngram_book", json=params)
    # print(r.status_code)
    df = pd.DataFrame.from_dict(r.json(), orient='index')
    df.index = df.index.map(lambda x: tuple(x.split()))
    columns = df.index.levels[0]
    df = pd.concat([df.loc[x] for x in columns], axis=1)
    df.columns = columns
    # df.index = df.index.map(pd.Timestamp)
    return df


# @_docstring_parameters_from(_ngram_doc, drop="doctype")
def ngram_periodicals(
        word: Union[List, str] = ["."],
        title: str = None,
        period: Tuple[int, int] = None,
        publisher: str = None,
        lang: str = None,
        city: str = None,
        ddk: str = None,
        topic: str = None,
        **kwargs,
) -> DataFrame:
    """Get a time series of frequency counts for ``word`` in periodicals.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/ngram_periodicals`.

    :param word: Word(s) to search for.
        Can be several words in a single string, separated by comma, e.g. ``"ord,ordene,orda"``.
    :type word: str or list of str
    :param str title: Title of a specific document to search through.
    :param period: Start and end years or dates of a time period,
        given as ``(YYYY, YYYY)`` or ``(YYYYMMDD, YYYYMMDD)``.
    :type period: tuple of ints
    :param str publisher: Name of a publisher.
    :param str lang: Language as a 3-letter ISO code (e.g. ``"nob"`` or ``"nno"``)
    :param str city: City of publication.
    :param str ddk: `Dewey Decimal Classification
        <https://no.wikipedia.org/wiki/Deweys_desimalklassifikasjon>`_ identifier.
    :param str topic: Topic of the documents.
    :return: a ``pandas.DataFrame`` with the resulting frequency counts of the word(s),
        spread across years. One year per row.
    """
    params = locals()
    if isinstance(word, str):
        # assume a comma separated string
        word = [w.strip() for w in word.split(',')]
    params['word'] = tuple(word)
    params = {x: params[x] for x in params if not params[x] is None}
    r = requests.post(BASE_URL + "/ngram_periodicals", json=params)
    # print(r.status_code)
    df = pd.DataFrame.from_dict(r.json(), orient='index')
    df.index = df.index.map(lambda x: tuple(x.split()))
    columns = df.index.levels[0]
    df = pd.concat([df.loc[x] for x in columns], axis=1)
    df.columns = columns
    # df.index = df.index.map(pd.Timestamp)
    return df


def ngram_news(
        word: Union[List, str] = ["."],
        title: str = None,
        period: Tuple[int, int] = None,
) -> DataFrame:
    """Get a time series of frequency counts for ``word`` in newspapers.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/ngram_newspapers`.

    :param word: Word(s) to search for.
        Can be several words in a single string, separated by comma, e.g. ``"ord,ordene,orda"``.
    :type word: str or list of str
    :param str title: Title of a specific newspaper to search through.
    :param period: Start and end years or dates of a time period,
        given as ``(YYYY, YYYY)`` or ``(YYYYMMDD, YYYYMMDD)``.
    :type period: tuple of ints
    :return: a ``pandas.DataFrame`` with the resulting frequency counts of the word(s),
        spread across the dates given in the time period. Either one year or one day per row.
    """
    params = locals()
    if isinstance(word, str):
        # assume a comma separated string
        word = [w.strip() for w in word.split(',')]
    params['word'] = tuple(word)
    params = {x: params[x] for x in params if not params[x] is None}
    r = requests.post(BASE_URL + "/ngram_newspapers", json=params)
    # print(r.status_code)
    df = pd.DataFrame.from_dict(r.json(), orient='index')
    df.index = df.index.map(lambda x: tuple(x.split()))
    columns = df.index.levels[0]
    df = pd.concat([df.loc[x] for x in columns], axis=1)
    df.columns = columns
    # df.index = df.index.map(pd.Timestamp)
    return df


def get_document_frequencies(urns: List[str] = None, cutoff: int = 0, words: List[str] = None) -> DataFrame:
    """Fetch frequency counts of ``words`` in documents (``urns``).

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/frequencies`.

    :param list urns: list of uniform resource name strings, for example:
        ``["URN:NBN:no-nb_digibok_2008051404065", "URN:NBN:no-nb_digibok_2010092120011"]``
    :param int cutoff: minimum frequency of a word to be counted
    :param list words: a list of words to be counted - if left None, whole document is returned. If not None both the counts and their relative frequency is returned.
    """
    params = locals()
    r = requests.post(f"{BASE_URL}/frequencies", json=params)
    result = r.json()
    # check if words are passed - return differs a bit
    if words is None:
        structure = dict()
        for u in result:
            try:
                structure[u[0][0]] = dict([(x[1], x[2]) for x in u])
            except IndexError:
                pass
        df = pd.DataFrame(structure)
        df = df.sort_values(by=df.columns[0], ascending=False).fillna(0)
    else:
        df = pd.DataFrame(result)
        df.columns = ["urn", "word", "freq", "urncount"]
        df['relfreq'] = df['freq']/df.urncount
        df = pd.pivot_table(df, values=["freq",'relfreq'], index="word", columns="urn").fillna(0)
    return df


def get_word_frequencies(
        urns: List[str] = None, cutoff: int = 0, words: List[str] = None
) -> DataFrame:
    """Fetch frequency numbers for ``words`` in documents (``urns``).

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/frequencies`.

    :param list urns: list of uniform resource name strings, for example:
        ``["URN:NBN:no-nb_digibok_2008051404065", "URN:NBN:no-nb_digibok_2010092120011"]``
    :param int cutoff: minimum frequency of a word to be counted
    :param list words: a list of words to be counted - should not be left None.
    """
    return get_document_frequencies(urns, cutoff, words)

def get_urn_frequencies(
        urns: List[str] = None, dhlabid: List = None
) -> DataFrame:
    """Fetch frequency counts of documents as URNs or DH-lab ids.
    
    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/frequencies`.

    :param list urns: list of uniform resource name strings, for example:
        ``["URN:NBN:no-nb_digibok_2008051404065", "URN:NBN:no-nb_digibok_2010092120011"]``
    :param list dhlabid: list of numbers for dhlabid:
        ``[1000001, 2000003]``
    """
    if dhlabid == None:
        params = {'urns': urns}
    else:
        params = {'dhlabid': dhlabid}
    r = requests.post(f"{BASE_URL}/urn_frequencies", json=params)
    result = r.json()
    # check if words are passed - return differs a bit
    df = pd.DataFrame(result)
    df.columns = ["urn", "freq"]
    return df

def get_document_corpus(**kwargs):
    return document_corpus(**kwargs)


def document_corpus(
        doctype: str = None,
        author: str = None,
        freetext: str = None,
        fulltext: str = None,
        from_year: int = None,
        to_year: int = None,
        from_timestamp: int = None,
        to_timestamp: int = None,
        title: str = None,
        ddk: str = None,
        subject: str = None,
        lang: str = None,
        limit: int = None,
        order_by: str = None,
) -> DataFrame:
    """Fetch a corpus based on metadata.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/build_corpus <https://api.nb.no/dhlab/#/default/post_build_corpus>`_.

    :param str doctype: ``"digibok"``, ``"digavis"``, ``"digitidsskrift"`` or ``"digistorting"``
    :param str author: Name of an author.
    :param str freetext: any of the parameters, for example: ``"digibok AND Ibsen"``.
    :param str fulltext: words within the publication.
    :param int from_year: Start year for time period of interest.
    :param int to_year: End year for time period of interest.
    :param int from_timestamp: Start date for time period of interest.
        Format: ``YYYYMMDD``, books have ``YYYY0101``
    :param int to_timestamp: End date for time period of interest.
        Format: ``YYYYMMDD``, books have ``YYYY0101``
    :param str title: Name or title of a document.
    :param str ddk: `Dewey Decimal Classification
        <https://no.wikipedia.org/wiki/Deweys_desimalklassifikasjon>`_ identifier.
    :param str subject: subject (keywords) of the publication.
    :param str lang: Language of the publication, as a 3-letter ISO code.
        Example: ``"nob"`` or ``"nno"``
    :param int limit: number of items to sample.
    :param str order_by: order of elements in the corpus object. Typically used in combination with a limit. Example ``"random"`` (random order, the slowest), ``"rank"`` (ordered by relevance, faster) or ``"first"`` (breadth-first, using the order in the database table, the fastest method) 
    :return: a ``pandas.DataFrame`` with the corpus information.
    """
    parms = locals()
    params = {x: parms[x] for x in parms if not parms[x] is None}

    r = requests.post(BASE_URL + "/build_corpus", json=params)

    return pd.DataFrame(r.json())


def urn_collocation(
        urns: List = None,
        word: str = "arbeid",
        before: int = 5,
        after: int = 0,
        samplesize: int = 200000,
) -> DataFrame:
    """Create a collocation from a list of URNs.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/urncolldist_urn`.

    :param list urns: list of uniform resource name strings, for example:
        ``["URN:NBN:no-nb_digibok_2008051404065", "URN:NBN:no-nb_digibok_2010092120011"]``
    :param str word: word to construct collocation with.
    :param int before: number of words preceding the given ``word``.
    :param int after: number of words following the given ``word``.
    :param int samplesize: total number of ``urns`` to search through.
    :return: a ``pandas.DataFrame`` with distance (sum of distances and bayesian distance) and
        frequency for words collocated with ``word``.
    """

    params = {
        "urn": urns,
        "word": word,
        "before": before,
        "after": after,
        "samplesize": samplesize,
    }
    r = requests.post(BASE_URL + "/urncolldist_urn", json=params)
    return pd.read_json(StringIO(r.json()))
    


def totals(top_words: int = 50000) -> DataFrame:
    """Get aggregated raw frequencies of all words in the National Library's database.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/totals/{top_words} <https://api.nb.no/dhlab/#/default/get_totals__top_words_>`_.

    :param int top_words: The number of words to get total frequencies for.
    :return: a ``pandas.DataFrame`` with the most frequent words.
    """
    r = requests.get(BASE_URL + f"/totals/{top_words}")
    return pd.DataFrame.from_dict(dict(r.json()), orient="index", columns=["freq"])


def concordance(
        urns: list = None, words: str = None, window: int = 25, limit: int = 100
) -> DataFrame:
    """Get a list of concordances from the National Library's database.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/conc <https://api.nb.no/dhlab/#/default/post_conc>`_.

    :param list urns: uniform resource names, for example:
        ``["URN:NBN:no-nb_digibok_2008051404065", "URN:NBN:no-nb_digibok_2010092120011"]``
    :param str words: Word(s) to search for.
        Can be an SQLite fulltext query, an fts5 string search expression.
    :param int window: number of tokens on either side to show in the collocations, between 1-25.
    :param int limit: max. number of concordances per document. Maximum value is 1000.
    :return: a table of concordances
    """
    if words is None:
        return {}  # exit condition
    else:
        params = {"urns": urns, "query": words, "window": window, "limit": limit}
        r = requests.post(BASE_URL + "/conc", json=params)
    return pd.DataFrame(r.json())


def concordance_counts(
        urns: list = None, words: str = None, window: int = 25, limit: int = 100
) -> DataFrame:
    """Count concordances (keyword in context) for a corpus query (used for collocation analysis).

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/conccount <https://api.nb.no/dhlab/#/default/post_conccount>`_.

    :param list urns: uniform resource names, for example:
        ``["URN:NBN:no-nb_digibok_2008051404065", "URN:NBN:no-nb_digibok_2010092120011"]``
    :param str words: Word(s) to search for.
        Can be an SQLite fulltext query, an fts5 string search expression.
    :param int window: number of tokens on either side to show in the collocations, between 1-25.
    :param int limit: max. number of concordances per document. Maximum value is 1000.
    :return: a table of counts
    """
    if words is None:
        return {}  # exit condition
    else:
        params = {"urns": urns, "query": words, "window": window, "limit": limit}
        r = requests.post(BASE_URL + "/conccount", json=params)
    return pd.DataFrame(r.json())


def konkordans(
        urns: list = None, words: str = None, window: int = 25, limit: int = 100
):
    """Wrapper for :func:`concordance`."""
    return concordance(**locals())


def word_concordance(
        urn: list = None, dhlabid: list = None, words: list = None, before: int = 12, after: int = 12, limit: int = 100, samplesize: int = 50000
) -> DataFrame:
    """Get a list of concordances from the National Library's database.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/conc <https://api.nb.no/dhlab/#/default/conc_word_urn>`_.

    :param list urns: dhlab serial ids. (server can take both urns and dhlabid but so we may rewrite this to)
    :param str words: Word(s) to search for -- must be a list
    :param int before: between 0-24.
    :param int after: between 0-24 (before + sum <= 24)
    :param int limit: max. number of concordances per server process.
    :param int samplesize: samples from urns.
    :return: a table of concordances
    """

    # server checks if either dhlabid or urns are present in the parameters, so only one of them
    # is passed. The return is dhlabid anyhow.

    if dhlabid is not None:
        params = {"dhlabid": dhlabid, "words": words, "before": before, "after": after, "limit": limit, "samplesize":samplesize}
    elif urns is not None:
        params = {"urn": urns, "words": words,  "before": before, "after": after, "limit": limit, "samplesize":samplesize}
    else:
        params = {"words": words,  "before": before, "after": after, "limit": limit, "samplesize":samplesize}

    r = requests.post(BASE_URL + "/conc_word_urn", json=params)

    return pd.DataFrame([x for y in r.json() for x in y], columns=['dhlabid', 'before','target','after'])


def collocation(
        corpusquery: str = "norge", word: str = "arbeid", before: int = 5, after: int = 0
) -> DataFrame:
    """Make a collocation from a corpus query.

    :param str corpusquery: query string
    :param str word: target word for the collocations.
    :param int before: number of words prior to ``word``
    :param int after: number of words following ``word``
    :return: a dataframe with the resulting collocations
    """
    params = {
        "metadata_query": corpusquery,
        "word": word,
        "before": before,
        "after": after,
    }
    r = requests.post(BASE_URL + "/urncolldist", json=params)
    return pd.read_json(r.json())


# Norwegian word bank


def word_variant(word: str, form: str, lang: str = "nob") -> list:
    """Find alternative ``form`` for a given ``word`` form.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint ``/variant_form``

    Example: ``word_variant('spiste', 'pres-part')``

    :param str word: any word string
    :param str form: a morphological feature tag from the Norwegian wordbank
        `"Orbanken" <https://www.nb.no/sprakbanken/ressurskatalog/oai-nb-no-sbr-5/>`_.
    :param str lang: either "nob" or "nno"
    """
    r = requests.get(
        f"{BASE_URL}/variant_form", params={"word": word, "form": form, "lang": lang}
    )
    return r.json()


def word_paradigm(word: str, lang: str = "nob") -> list:
    """Find paradigms for a given ``word`` form.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint ``/paradigm``

    Example:

    .. code-block:: python

        word_paradigm('spiste')
        # [['adj', ['spisende', 'spist', 'spiste']],
        # ['verb', ['spis', 'spise', 'spiser', 'spises', 'spist', 'spiste']]]

    :param str word: any word string
    :param str lang: either "nob" or "nno"
    """
    r = requests.get(f"{BASE_URL}/paradigm", params={"word": word, "lang": lang})
    return r.json()


def word_paradigm_many(wordlist: list, lang: str = "nob") -> list:
    """Find alternative forms for a list of words."""
    r = requests.post(f"{BASE_URL}/paradigms", json={"words": wordlist, "lang": lang})
    return r.json()


def word_form(word: str, lang: str = "nob") -> list:
    """Look up the morphological feature specification of a ``word`` form."""
    r = requests.get(f"{BASE_URL}/word_form", params={"word": word, "lang": lang})
    return r.json()


def word_form_many(wordlist: list, lang: str = "nob") -> list:
    """Look up the morphological feature specifications for word forms in a ``wordlist``."""
    r = requests.post(f"{BASE_URL}/word_forms", json={"words": wordlist, "lang": lang})
    return r.json()


def word_lemma(word: str, lang: str = "nob") -> list:
    """Find the list of possible lemmas for a given ``word`` form."""
    r = requests.get(f"{BASE_URL}/word_lemma", params={"word": word, "lang": lang})
    return r.json()


def word_lemma_many(wordlist, lang="nob"):
    """Find lemmas for a list of given word forms."""

def query_imagination_corpus(category=None, author=None, title=None, year=None, publisher=None, place=None, oversatt=None):
    """Fetch data from imagination corpus"""
    params = locals()
    params = {key: params[key] for key in params if params[key] is not None}
    #print(params)
    r = requests.get(f"{BASE_URL}/imagination", params=params)
    return r.json()



from typing import Union

import pandas as pd

from dhlab.api.dhlab_api import (
    word_form,
    word_form_many,
    word_lemma,
    word_lemma_many,
    word_paradigm,
    word_paradigm_many,
    word_variant,
)


class WordbankSuper():
    """Super class for wordbank classes"""
    def __init__(self, frame):
        self.frame = frame
    
    def __repr__(self) -> str:
        return self.frame.__repr__()
    
    def _repr_html_(self) -> Union[str, None]:
        """
        Return the HTML representation of the DhlabObj frame attribute
        """
        return self.frame._repr_html_()


class WordParadigm(WordbankSuper):
    """Fetch inflection paradigms for a list of words, or just one word"""
    def __init__(self, words, lang='nob'):
        if isinstance(words, list):
            self.paradigms = pd.DataFrame(word_paradigm_many(words, lang))
        elif isinstance(words, str):
            self.paradigms = pd.DataFrame(word_paradigm(words, lang))
        else:
            self.paradigms = None
        
        super().__init__(self.paradigms)
            
class WordForm(WordbankSuper):
    """Fetch possible forms of a word or list of words"""
    def __init__(self, words, lang='nob'):
        if isinstance(words, list):
            self.forms = pd.DataFrame(word_form_many(words, lang))
        elif isinstance(words, str):
            self.forms = pd.DataFrame(word_form(words, lang))
        else:
            self.forms = None

        super().__init__(self.forms)
        
class WordLemma(WordbankSuper):
    """Fetch possbile lemmas for a given word form"""
    def __init__(self, words, lang='nob'):
        if isinstance(words, list):
            self.lemmas = pd.DataFrame(word_lemma_many(words, lang))
        elif isinstance(words, str):
            self.lemmas = pd.DataFrame(word_lemma(words, lang))
        else:
            self.lemmas = None
            
        super().__init__(self.lemmas)
    
    


from datetime import datetime

import pandas as pd

from dhlab.api.dhlab_api import ngram_book, ngram_news
from dhlab.ngram.nb_ngram import nb_ngram


class Ngram(pd.DataFrame):
    """A sketch for a new Ngram class."""

    _metadata = ["from_year", "to_year"]

    @property
    def _constructor(self):
        return Ngram

    @classmethod
    def build(
        self,
        words=None,
        from_year=1950,
        to_year=datetime.now().year,
        doctype="bok",
        mode="relative",
        lang="nob",
        **kwargs
    ):
        res = nb_ngram(
            terms=", ".join(words),
            corpus=doctype,
            years=(from_year, to_year),
            lang=lang,
            mode=mode,
        )

        return Ngram(res)

    @classmethod
    def from_news(
        self,
        words=None,
        title=None,
        city=None,
        from_year=1950,
        to_year=datetime.now().year,
        **kwargs
    ):
        res = ngram_news(word=words, period=(from_year, to_year))

        return Ngram(res)

    @classmethod
    def from_book(
        self,
        words=None,
        title=None,
        publisher=None,
        city=None,
        lang="nob",
        from_year=1950,
        to_year=datetime.now().year,
        ddk=None,
        subject=None,
        **kwargs
    ):
        res = ngram_book(
            word=words,
            title=title,
            publisher=publisher,
            lang=lang,
            city=city,
            period=(from_year, to_year),
            ddk=ddk,
            topic=subject,
        )

        self.from_year = from_year
        self.to_year = to_year

        return Ngram(res)

    def ngram_plot(self, smooth=4, **kwargs):
        grf = self.rolling(window=smooth, win_type="triang").mean()
        grf.plot(**kwargs)

    def compare(self, another_ngram):
        """Divide one ngram by another - measures difference"""
        start_year = max(
            datetime(self.from_year, 1, 1), datetime(another_ngram.from_year, 1, 1)
        ).year
        end_year = min(
            datetime(self.to_year, 1, 1), datetime(another_ngram.to_year, 1, 1)
        ).year
        transposed_ngram = self.loc[str(start_year) : str(end_year)].transpose()
        sum_other_ngram = (
            another_ngram[str(start_year) : str(end_year)].transpose().sum()
        )
        compare = (transposed_ngram / sum_other_ngram).transpose()
        return compare



import json

import networkx as nx
import requests

from dhlab.constants import GALAXY_API, NGRAM_API


def get_ngram(terms: str, corpus: str = "avis", lang: str = "nob") -> dict:
    """Fetch raw and relative frequencies for the ``terms``.

    Call the :py:data:`NGRAM_API`.
    The frequencies are aggregated per year between 1800-2021.

    :param str terms: comma separated string of words
    :param str corpus: type of documents to search through
    :return: table of annual frequency counts per term
    """
    req = requests.get(
        NGRAM_API,
        params={
            'terms': terms,
            'corpus': corpus,
            'lang':lang
        }
    )
    if req.status_code == 200:
        res = req.text
    else:
        res = "[]"
    return json.loads(res)


def make_word_graph(
        words: str, corpus: str = 'all', cutoff: int = 16, leaves: int = 0
) -> nx.DiGraph:
    """Get galaxy from ngram-database.

    Call the :py:obj:`~dhlab.constants.GALAXY_API` endpoint.

    :param str words: comma-separated string of words
    :param str corpus: document type: ``'book'``, ``'avis'``, or ``'all'``,
    :param int cutoff: Number of nodes to include.
    :param int leaves: Set leaves=1 to get the leaves.
    :return: A `networkx.DiGraph` with the results.
    """

    params = dict()
    params['terms'] = words
    params['corpus'] = corpus
    params['limit'] = cutoff
    params['leaves'] = leaves
    result = requests.get(GALAXY_API, params=params)
    G = nx.DiGraph()
    edgelist = []
    if result.status_code == 200:
        graph = json.loads(result.text)
        # print(graph)
        nodes = graph['nodes']
        edges = graph['links']
        for edge in edges:
            edgelist += [
                (nodes[edge['source']]['name'],
                 nodes[edge['target']]['name'],
                 abs(edge['value']))
            ]
    # print(edgelist)
    G.add_weighted_edges_from(edgelist)
    return G



import re

import pandas as pd

import dhlab as dh
from dhlab.api.dhlab_api import concordance, get_document_frequencies, urn_collocation
from dhlab.text.dhlab_object import DhlabObj
from dhlab.text.utils import urnlist


# convert cell to a link
def make_link(row):
    r = "<a target='_blank' href = 'https://urn.nb.no/{x}'>{x}</a>".format(x=str(row))
    return r


# find hits a cell
def find_hits(x):
    return " ".join(re.findall("<b>(.+?)</b", x))


class Concordance(DhlabObj):
    """Wrapper for concordance function"""

    def __init__(self, corpus=None, query=None, window=20, limit=500):
        """Get concordances for word(s) in corpus

        :param corpus: Target corpus, defaults to None
        :param query: word or list or words, defaults to None
        :param window: how many tokens to consider around the target word, \
            defaults to 20
        :param limit: limit returned hits, defaults to 500
        """

        if corpus is None:
            self.concordance = pd.DataFrame()
            self.corpus = None
        else:
            self.concordance = concordance(
                urns=urnlist(corpus), words=query, window=window, limit=limit
            )
            self.concordance["link"] = self.concordance.urn.apply(make_link)
            self.concordance = self.concordance[["link", "urn", "conc"]]
            self.concordance.columns = ["link", "urn", "concordance"]
            self.corpus = corpus
        # self.size = len(self.concordance)

        super().__init__(self.concordance)

    def show(self, n=10, style=True):
        if style:
            result = self.concordance.sample(min(n, self.size))[
                ["link", "concordance"]
            ].style
        else:
            result = self.concordance.sample(min(n, self.size))
        return result
    
    @classmethod
    def from_df(cls, df):
        "Typecast DataFrame to Concordance"
        obj = Concordance()
        obj.concordance = df
        obj.frame = df
        return obj


class Collocations(DhlabObj):
    """Collocations"""

    def __init__(
        self,
        corpus=None,
        words=None,
        before=10,
        after=10,
        reference=None,
        samplesize=20000,
        alpha=False,
        ignore_caps=False,
    ):
        """Create collocations object

        :param corpus: target corpus, defaults to None
        :type corpus: dh.Corpus, optional
        :param words: target words(s), defaults to None
        :type words: str or list, optional
        :param before: words to include before, defaults to 10
        :type before: int, optional
        :param after: words to include after, defaults to 10
        :type after: int, optional
        :param reference: reference frequency list, defaults to None
        :type reference: pd.DataFrame, optional
        :param samplesize: _description_, defaults to 20000
        :type samplesize: int, optional
        :param alpha: Only include alphabetical tokens, defaults to False
        :type alpha: bool, optional
        :param ignore_caps: Ignore capitalized letters, defaults to False
        :type ignore_caps: bool, optional
        """
        if isinstance(words, str):
            words = [words]

        coll = pd.concat(
            [
                urn_collocation(
                    urns=urnlist(corpus),
                    word=w,
                    before=before,
                    after=after,
                    samplesize=samplesize,
                )
                for w in words
            ]
        )[["counts"]]

        if alpha:
            coll = coll.loc[[x for x in coll.index if x.isalpha()]]
            if reference is not None:
                reference = reference.loc[[x for x in reference.index if x.isalpha()]]

        if ignore_caps:
            coll.index = [x.lower() for x in coll.index]
            if reference is not None:
                reference.index = [x.lower() for x in reference.index]

        self.coll = coll.groupby(coll.index).sum()
        self.reference = reference
        self.before = before
        self.after = after

        if reference is not None:
            teller = self.coll.counts / self.coll.counts.sum()
            divisor = self.reference.iloc[:, 0] / self.reference.iloc[:, 0].sum()
            self.coll["relevance"] = teller / divisor

        super().__init__(self.coll)

    def show(self, sortby="counts", n=20):
        return self.coll.sort_values(by=sortby, ascending=False).head(n)

    def keywordlist(self, top=200, counts=5, relevance=10):
        mask = self.coll[self.coll.counts > counts]
        mask = mask[mask.relevance > relevance]
        return list(mask.sort_values(by="counts", ascending=False).head(200).index)

    @classmethod
    def from_df(cls, df):
        """Typecast DataFrame to Collocation

        :param df: DataFrame
        :return: Collocation
        """
        obj = Counts()
        obj.counts = df
        obj.frame = df
        return obj


class Counts(DhlabObj):
    """Provide counts for a corpus - shouldn't be too large"""

    def __init__(self, corpus=None, words=None):
        """Get frequency list for Corpus

        :param corpus: target Corpus, defaults to None
        :param words: list of words to be counted, defaults to None
        """
        if corpus is None and words is None:
            self.freq = pd.DataFrame()
            self.title_dct = None
        elif corpus is not None:
            # Make sure corpus is a dhlab corpus
            # if not, try to make it one
            # if isinstance(corpus, pd.DataFrame):
            #     corpus = dh.Corpus.from_df(corpus)
            
            # if not isinstance(corpus, dh.Corpus):
            #     raise TypeError("Corpus must be of type dh.Corpus or pd.DataFrame")

            # count - if words is none result will be as if counting all words
            # in the corpus
            self.freq = get_document_frequencies(
                urns=urnlist(corpus), cutoff=0, words=words
            )

            # Include dhlab and title link in object
            try:
                self.title_dct = {
                    k: v for k, v in zip(corpus.frame.dhlabid, corpus.frame.title)
                }
            except:
                self.title_dct = None

            # Add relative frequencies if available
            if words is not None:
                self.relfreq = self.freq.relfreq
                self.freq = self.freq.freq

        super().__init__(self.freq)

    def sum(self):
        """Summarize Corpus frequencies

        :return: frequency list for Corpus
        """
        return self.from_df(self.counts.sum(axis=1).to_frame("freq"))

    def display_names(self):
        "Display data with record names as column titles."
        assert self.title_dct is not None, "No titles available"            
        return self.frame.rename(self.title_dct, axis=1)
    
    def display_rel_names(self):
        "Display relfreq data with record names as column titles."
        return self.relfreq.rename(self.title_dct, axis=1)

    @classmethod
    def from_df(cls, df):
        obj = Counts()
        obj.freq = df
        obj.frame = df
        return obj

    ### Legacy properties and methods ###

    @property
    def counts(self):
        "Legacy property for freq"
        return self.freq




import pandas as pd

from dhlab.api.dhlab_api import geo_lookup, get_places, ner_from_urn
from dhlab.text.parse import NER, Models


class GeoData:
    """Fetch place data from a text (book, newspaper or ...) identified by URN
    with an appropriate and available spacy model.

    The models are retrieved by instantiating :py:class:`~text.parse.Models`.
    """
    
    def __init__(self, urn=None, model=None):
        self.place_names = self._fetch_place_names(urn, model)
        

    def _fetch_place_names(self, urn, model):
        try:
            assert urn is not None
            spacy = Models()
            model = model if model is not None else spacy.models[0]
            df = NER(urn = urn, model = model).ner
            place_names = df[df['ner'].str.contains('LOC')]
        except AssertionError:
            print("Please provide a document URN to fill the ``place_names`` dataframe attribute.")
            place_names = pd.DataFrame()
        except IndexError as error:
            print("GeoData couldn't load any SpaCy NER models.")
            place_names = pd.DataFrame()
        except Exception as error:
            print(error.__doc__, error)
            place_names = pd.DataFrame()
        return place_names


    def add_geo_locations(self, feature_class = None, feature_code = None):
        """Get location data for the names in object, attribute self.place_names"""
        chunksize = 900
        names = list(self.place_names.token)
        length = len(names)
        
        # GeoNames takes 1000 names at a time so chunk things up
        # Each GeoNames object has an attribute .places a pandas dataframe
        
        chunks = [
            GeoNames(
                names[i:i+chunksize], 
                feature_class = feature_class, 
                feature_code = feature_code
            ).places for i in range(0, length, chunksize)
        ] 
        
        self.places = pd.concat(chunks) 

        
class GeoNames:
    """Fetch data from a list of names"""
    def __init__(self, names, feature_class = None, feature_code = None):
        self.places = geo_lookup(names, feature_class = feature_class, feature_code = feature_code)
        
        


"""Constant variables used across modules."""

BASE_URL = "https://api.nb.no/dhlab"    #: REST-API URL adress to fulltext query functions
NGRAM_API = "https://api.nb.no/dhlab/nb_ngram/ngram/query"  #: URL adress for API calls to ngram-databases
GALAXY_API = "https://api.nb.no/dhlab/nb_ngram_galaxies/galaxies/query"     #: URL adress for word galaxy API queries



from pandas import DataFrame, Series

import dhlab as dh


def urnlist(corpus):
    """Try to pull out a list of URNs from corpus"""
    if isinstance(corpus, dh.Corpus):
        _urnlist = list(corpus.corpus.urn)
    elif isinstance(corpus, DataFrame):
        _urnlist = list(corpus.urn)
    elif isinstance(corpus, list):
        _urnlist = corpus
    elif isinstance(corpus, Series):
        _urnlist = corpus.to_list()
    else:
        _urnlist = []
    return _urnlist



from typing import Iterable

import pandas as pd
import requests


def load_picture(url: str):
    """Load the raw image object from a URL."""
    r = requests.get(url, stream=True)
    r.raw.decode_content = True
    # print(r.status_code)
    return r.raw


def iiif_manifest(urn: str):
    """Fetch the IIIF manifest of the images that the book """
    r = requests.get(f"https://api.nb.no/catalog/v1/iiif/{urn}/manifest")
    return r.json()


def mods(urn: str):
    r = requests.get(f"https://api.nb.no:443/catalog/v1/metadata/{urn}/mods")
    return r.json()


def super_search(term: str, number: int = 50, page: int = 0, mediatype: str = 'bilder') -> dict:
    """Search for a term, return a json object with the matches.

    :param str term: Search term, word, or token. Default is the empty string.
    :param int number: Size of the response. Maximum 50.
    :param int page: The page number to search from.
        .. todo:: Fill in description.
    :param str mediatype: Default 'bilder'.
        .. todo::  Which other value options are available for this parameter?
    :return: a dict (json object) with the search results.
    """
    url = "https://api.nb.no:443/catalog/v1/items"
    number = min(number, 50)
    if term == '':
        r = requests.get(
            url,
            params={
                'filter': f'mediatype:{mediatype}',
                'page': page,
                'size': number
            }
        )
    else:
        r = requests.get(
            url,
            params={
                'q': term,
                'filter': f'mediatype:{mediatype}',
                'page': page,
                'size': number
            }
        )
    return r.json()


def total_search(size: int = 50, page: int = 0) -> dict:
    """Retrieve the first ``size`` occurrences from ``page``.

    Wrapper-function for :func:`super_search` with arguments
    ``medietype="bilder"`` and ``term=''``.
    """
    return super_search(term='', number=size, page=page, mediatype='bilder')


def get_df(phrases: Iterable, title: str = 'aftenposten'):
    """Get ``phrases`` from documents of ``title``, and aggregate their frequencies."""
    querystring = " + ".join(['"' + frase + '"' for frase in phrases])
    query = {
        'q': querystring,
        'size': 1,
        'aggs': 'year',
        # 'filter':'mediatype:{mt}'.format(mt=media),
        'filter': f'title:{title}'
    }
    r = requests.get("https://api.nb.no/catalog/v1/items", params=query)
    aggs = r.json()['_embedded']['aggregations'][0]['buckets']
    return {x['key']: x['count'] for x in aggs}


def get_json(phrases, mediatype='aviser'):
    querystring = " + ".join(['"' + frase + '"' for frase in phrases])
    query = {
        'q': querystring,
        'size': 1,
        'snippets': mediatype,
        'aggs': 'year',

        #        'filter':'mediatype:{mt}'.format(mt=mediatype),
        'searchType': 'FULL_TEXT_SEARCH'
        # 'filter':'title:{title}'.format(title=title)
    }
    r = requests.get("https://api.nb.no/catalog/v1/items", params=query)
    aggs = r.json()
    return aggs


def get_data(frase, media='avis', title='jazznytt'):

    query = {
        'q': '"' + frase + '""',
        'size': 1,
        'aggs': 'year',
        'filter': 'mediatype:{mt}'.format(mt=media),
        'filter': 'title:{title}'.format(title=title)
    }
    r = requests.get("https://api.nb.no/catalog/v1/items", params=query)
    return r.json()


def get_data_and(frases, title='aftenposten', media='avis'):

    querystring = " + ".join(['"' + frase + '"' for frase in frases])
    print(querystring)
    query = {
        'q': querystring,
        'size': 1,
        'aggs': 'year',
        # 'filter':'mediatype:{mt}'.format(mt=media),
        'filter': 'title:{title}'.format(title=title)
    }
    r = requests.get("https://api.nb.no/catalog/v1/items", params=query)
    return r.json()


def get_df_pd(frase, media='bøker'):
    return pd.DataFrame.from_dict(
        get_df(frase, media=media), orient='index').sort_index()


def get_konks(urn, phrase, window=1000, n=1000):

    querystring = '"' + phrase + '"'
    query = {
        'q': querystring,
        'fragments': n,
        'fragSize': window

    }
    r = requests.get(
        "https://api.nb.no/catalog/v1/items/{urn}/contentfragments".format(urn=urn), params=query)
    res = r.json()
    results = []
    try:
        for x in res['contentFragments']:
            urn = x['pageid']
            hit = x['text']
            splits = hit.split('<em>')
            s2 = splits[1].split('</em>')
            before = splits[0]
            word = s2[0]
            after = s2[1]
            results.append({'urn': urn, 'before': before,
                           'word': word, 'after': after})
    except BaseException:
        results = []
    return results


def get_phrase_info(urn, phrase, window=1000, n=1000):

    querystring = '"' + phrase + '"'
    query = {
        'q': querystring,

    }
    r = requests.get(
        f"https://api.nb.no/catalog/v1/items/{urn}/contentfragments",
        params=query)
    res = r.json()
    return res



from dhlab.api.dhlab_api import wildcard_search
from dhlab.text.dhlab_object import DhlabObj


class WildcardWordSearch(DhlabObj):
    """
    Find a class of words matching a wildcard string
    """
    def __init__(self, word, factor=2, freq_limit=10, limit=50):
        """:param word: word from a mixture of * and characters
        :factor int: the additional length of words to be returned
        :freq_lim: the frequency of returned words lower limit
        :limit int: number of words returned
        """
        self.words = wildcard_search(word, factor=factor,
                                     freq_limit=freq_limit, limit=limit)
        super().__init__(self.words)



"""Utility functions."""
import re


def _is_documented_by(original):
    """Decorator to reuse docstrings for several functions.

    Fetch the docstring of another ``original`` function,
    and assign it as the docstring of the decorated function.

    Code copied from `StackExchange <https://softwareengineering.stackexchange.com/a/386758>`_.

    :param Callable original: Function that is decorated
    :return: The ``target`` function, with the ``original`` docstring

    **Example use:**

    .. code-block:: python

       def original(args):
           '''A complete docstring.

           :param args: Arguments.
           '''
           ...

       @_is_documented_by(original)
       def target(args):
           ...
    """
    def wrapper(target):
        target.__doc__ = original.__doc__
        return target
    return wrapper


def _docstring_parameters_from(original, drop: str = None):
    """Extract only parameter descriptions from ``original``'s docstring.

    Optionally, specify a ``drop`` parameter to skip in the parameter list,
    and append the rest. Can be a comma-separated string of several parameters.

    Otherwise, the wrapper works the same way as :func:`_is_documented_by`.
    """

    def _remove_param_desc(text, skip):
        regx = re.compile(fr'(\s*:param.+{skip}:\s.+\n\s *.+)(?=\n\s*:param)')
        skip_str = re.search(regx, text).group()
        return text.replace(skip_str, "")

    def wrapper(target):
        docstr = original.__doc__
        parameters = docstr[docstr.index("\n\n    :param"):]
        if drop is not None:
            for skip in drop.split(","):
                parameters = _remove_param_desc(parameters, skip)
        target.__doc__ += "\n" + parameters
        return target
    return wrapper



from abc import ABC, abstractmethod
from typing import Union

import pandas as pd


class DhlabObj(ABC):
    """DHLAB base class

    Provides shared utility methods to DHLAB classes.
    """

    def __init__(self, frame: pd.DataFrame = pd.DataFrame()):
        self.frame = frame

        # self.size = None
        # if isinstance(frame, pd.DataFrame):
        #     self.size = len(frame)

    @property
    def size(self):
        "Return number of rows in dataframe"
        return len(self.frame)

    @property
    def loc(self):
        return self.frame.loc

    def make_subset(self, row_slice, col_slice):
        "Return a subset of the dataframe"
        return self.from_df(self.frame.loc[:, col_slice].iloc[row_slice])

    def __getitem__(self, *args, **kwargs):
        "Return a subset of the dataframe"
        return self.from_df(self.frame.__getitem__(*args, **kwargs))

    def __repr__(self) -> str:
        """
        Return the string representation of the  DhlabObj frame attribute
        """
        return self.frame.__repr__()

    def _repr_html_(self) -> Union[str, None]:
        """
        Return the HTML representation of the DhlabObj frame attribute
        """
        return self.frame._repr_html_()
    
    def __len__(self):
        return len(self.frame)

    def head(self, n=5):
        "Return first n rows"
        return self.from_df(self.frame.head(n))

    def tail(self, n=5):
        "Return last n rows"
        return self.from_df(self.frame.tail(n))

    def sort(self, by=None, asc=False):
        "Sort by column"
        if by is None:
            by = self.frame.columns[0]
        return self.from_df(self.frame.sort_values(by=by, ascending=asc))

    def to_csv(self, path):
        "Write to csv"
        self.frame.to_csv(path, index=None)

    def to_excel(self, path):
        "Write to excel"
        self.frame.to_excel(path, index=None)

    @abstractmethod
    def from_df(cls, df):
        "Typecast Pandas DataFrame to dhlab class"
        pass

    @classmethod
    def from_csv(cls, path):
        "Import corpus from csv"
        df = pd.read_csv(path)
        return cls.from_df(df)
